@@ -27,6 +27,7 @@
 #include "input.h"
 #include "callbacks.h"
 #include "config.h"
+#include "build-config.h"
 
 
 gboolean on_expose (GtkWidget *widget,
@@ -126,9 +127,9 @@ void on_monitors_changed ( GdkScreen *screen,
 
 
   data->default_pen = paint_context_new (data, GROMIT_PEN,
-					 data->red, 7, 0);
+					 data->red, 7, 0, 1);
   data->default_eraser = paint_context_new (data, GROMIT_ERASER,
-					    data->red, 75, 0);
+					    data->red, 75, 0, 1);
 
   if(!data->composited) // set shape
     {
@@ -160,7 +161,7 @@ void on_composited_changed ( GdkScreen *screen,
       // undo shape
       gtk_widget_shape_combine_region(data->win, NULL);
       // re-apply transparency
-      gtk_window_set_opacity(GTK_WINDOW(data->win), 0.75);
+      gtk_widget_set_opacity(data->win, 0.75);
     }
 
   // set anti-aliasing
@@ -231,7 +232,7 @@ gboolean on_buttonpress (GtkWidget *win,
 			 gpointer user_data)
 {
   GromitData *data = (GromitData *) user_data;
-  gdouble pressure = 0.5;
+  gdouble pressure = 1;
 
   /* get the data for this device */
   GdkDevice *master = ev->device;
@@ -261,16 +262,14 @@ gboolean on_buttonpress (GtkWidget *win,
   slavedata->lasty = ev->y;
   slavedata->motion_time = ev->time;
 
-  if (gdk_device_get_source(slave) == GDK_SOURCE_MOUSE)
-    {
-      data->maxwidth = slavedata->cur_context->width;
-    }
-  else
-    {
-      gdk_event_get_axis ((GdkEvent *) ev, GDK_AXIS_PRESSURE, &pressure);
-      data->maxwidth = (CLAMP (pressure * pressure + line_thickener,0,1) *
-                        (double) slavedata->cur_context->width);
-    }
+  snap_undo_state (data);
+
+  gdk_event_get_axis ((GdkEvent *) ev, GDK_AXIS_PRESSURE, &pressure);
+  data->maxwidth = (CLAMP (pressure + line_thickener, 0, 1) *
+		    (double) (slavedata->cur_context->width -
+			      slavedata->cur_context->minwidth) +
+		    slavedata->cur_context->minwidth);
+
   if (ev->button <= 5)
     draw_line (data, slave, ev->x, ev->y, ev->x, ev->y);
 
@@ -288,7 +287,7 @@ gboolean on_motion (GtkWidget *win,
   GdkTimeCoord **coords = NULL;
   gint nevents;
   int i;
-  gdouble pressure = 0.5;
+  gdouble pressure = 1;
 
   /* get the data for this device */
   GdkDevice *master = ev->device;
@@ -303,6 +302,9 @@ gboolean on_motion (GtkWidget *win,
   GromitDeviceData *slavedata =
     g_hash_table_lookup(data->devdatatable, slave);
 
+  if(data->debug)
+      g_printerr("DEBUG: Device '%s': motion to (x,y)=(%.2f : %.2f)\n", gdk_device_get_name(slave), ev->x, ev->y);
+
   if (ev->state != masterdata->state ||
       ev->state != slavedata->state ||
       masterdata->lastslave != slave)
@@ -329,11 +331,10 @@ gboolean on_motion (GtkWidget *win,
                                GDK_AXIS_PRESSURE, &pressure);
           if (pressure > 0)
             {
-              if (gdk_device_get_source(slave) == GDK_SOURCE_MOUSE)
-                data->maxwidth = slavedata->cur_context->width;
-              else
-                data->maxwidth = (CLAMP (pressure * pressure + line_thickener, 0, 1) *
-                                  (double) slavedata->cur_context->width);
+	      data->maxwidth = (CLAMP (pressure + line_thickener, 0, 1) *
+				(double) (slavedata->cur_context->width -
+					  slavedata->cur_context->minwidth) +
+				slavedata->cur_context->minwidth);
 
               gdk_device_get_axis(slave, coords[i]->axes,
                                   GDK_AXIS_X, &x);
@@ -357,11 +358,10 @@ gboolean on_motion (GtkWidget *win,
 
   if (pressure > 0)
     {
-      if (gdk_device_get_source(slave) == GDK_SOURCE_MOUSE)
-	data->maxwidth = slavedata->cur_context->width;
-      else
-        data->maxwidth = (CLAMP (pressure * pressure + line_thickener,0,1) *
-                          (double)slavedata->cur_context->width);
+      data->maxwidth = (CLAMP (pressure + line_thickener, 0, 1) *
+			(double) (slavedata->cur_context->width -
+				  slavedata->cur_context->minwidth) +
+			slavedata->cur_context->minwidth);
 
       if(slavedata->motion_time > 0)
 	{
@@ -421,7 +421,6 @@ gboolean on_buttonrelease (GtkWidget *win,
   return TRUE;
 }
 
-
 /* Remote control */
 void on_mainapp_selection_get (GtkWidget          *widget,
 			       GtkSelectionData   *selection_data,
@@ -432,22 +431,27 @@ void on_mainapp_selection_get (GtkWidget          *widget,
   GromitData *data = (GromitData *) user_data;
   
   gchar *uri = "OK";
+  GdkAtom action = gtk_selection_data_get_target(selection_data);
 
-  if(gtk_selection_data_get_target(selection_data) == GA_TOGGLE)
+  if(action == GA_TOGGLE)
     {
       /* ask back client for device id */
       gtk_selection_convert (data->win, GA_DATA,
                              GA_TOGGLEDATA, time);
       gtk_main(); /* Wait for the response */
     }
-  else if (gtk_selection_data_get_target(selection_data) == GA_VISIBILITY)
+  else if (action == GA_VISIBILITY)
     toggle_visibility (data);
-  else if (gtk_selection_data_get_target(selection_data) == GA_CLEAR)
+  else if (action == GA_CLEAR)
     clear_screen (data);
-  else if (gtk_selection_data_get_target(selection_data) == GA_RELOAD)
+  else if (action == GA_RELOAD)
     setup_input_devices(data);
-  else if (gtk_selection_data_get_target(selection_data) == GA_QUIT)
+  else if (action == GA_QUIT)
     gtk_main_quit ();
+  else if (action == GA_UNDO)
+    undo_drawing (data);
+  else if (action == GA_REDO)
+    redo_drawing (data);
   else
     uri = "NOK";
 
@@ -545,32 +549,48 @@ void on_device_added (GdkDeviceManager *device_manager,
 
 
 
+gboolean on_toggle_paint(GtkWidget *widget,
+			 GdkEventButton  *ev,
+			 gpointer   user_data)
+{
+    GromitData *data = (GromitData *) user_data;
 
-static void on_clear (GtkMenuItem *menuitem,
-		      gpointer     user_data)
+    GdkDevice *master = ev->device;
+    GdkDevice *slave = gdk_event_get_source_device ((GdkEvent *) ev);
+
+    if(data->debug)
+	g_printerr("DEBUG: Device '%s': Button %i on_toggle_paint at (x,y)=(%.2f : %.2f)\n",
+		   gdk_device_get_name(slave), ev->button, ev->x, ev->y);
+
+    toggle_grab(data, master);
+
+    return TRUE;
+}
+
+void on_clear (GtkMenuItem *menuitem,
+	       gpointer     user_data)
 {
   GromitData *data = (GromitData *) user_data;
   clear_screen(data);
 }
 
 
-static void on_toggle_vis(GtkMenuItem *menuitem,
-			  gpointer     user_data)
+void on_toggle_vis(GtkMenuItem *menuitem,
+		   gpointer     user_data)
 {
   GromitData *data = (GromitData *) user_data;
   toggle_visibility(data);
 }
 
 
-
-static void on_thicker_lines(GtkMenuItem *menuitem,
-			     gpointer     user_data)
+void on_thicker_lines(GtkMenuItem *menuitem,
+		      gpointer     user_data)
 {
   line_thickener += 0.1;
 }
 
-static void on_thinner_lines(GtkMenuItem *menuitem,
-			     gpointer     user_data)
+void on_thinner_lines(GtkMenuItem *menuitem,
+		      gpointer     user_data)
 {
   line_thickener -= 0.1;
   if (line_thickener < -1)
@@ -578,101 +598,89 @@ static void on_thinner_lines(GtkMenuItem *menuitem,
 }
 
 
-void on_trayicon_activate (GtkStatusIcon *status_icon,
-			   gpointer       user_data)
+void on_opacity_bigger(GtkMenuItem *menuitem,
+		       gpointer     user_data)
 {
   GromitData *data = (GromitData *) user_data;
-  if(data->debug)
-    g_printerr("DEBUG: trayicon activated\n");
+  data->opacity += 0.1;
+  if(data->opacity>1.0)
+    data->opacity = 1.0;
+  gtk_widget_set_opacity(data->win, data->opacity);
+}
 
-  /* create the menu */
-  GtkWidget *menu = gtk_menu_new ();
-  
-  /* Create the menu items */
-  GtkWidget* toggle_paint_item = gtk_image_menu_item_new_with_label ("Toggle Painting");
-  GtkWidget* clear_item = gtk_image_menu_item_new_with_label ("Clear Screen");
-  GtkWidget* toggle_vis_item = gtk_image_menu_item_new_with_label ("Toggle Visibility");
-  GtkWidget* thicker_lines_item = gtk_image_menu_item_new_with_label ("Thicker Lines");
-  GtkWidget* thinner_lines_item = gtk_image_menu_item_new_with_label ("Thinner Lines");
+void on_opacity_lesser(GtkMenuItem *menuitem,
+		       gpointer     user_data)
+{
+  GromitData *data = (GromitData *) user_data;
+  data->opacity -= 0.1;
+  if(data->opacity<0.0)
+    data->opacity = 0.0;
+  gtk_widget_set_opacity(data->win, data->opacity);
+}
 
 
-  /* Add them to the menu */
-  gtk_menu_shell_append (GTK_MENU_SHELL (menu), toggle_paint_item);
-  gtk_menu_shell_append (GTK_MENU_SHELL (menu), clear_item);
-  gtk_menu_shell_append (GTK_MENU_SHELL (menu), toggle_vis_item);
-  gtk_menu_shell_append (GTK_MENU_SHELL (menu), thicker_lines_item);
-  gtk_menu_shell_append (GTK_MENU_SHELL (menu), thinner_lines_item);
+void on_undo(GtkMenuItem *menuitem,
+	     gpointer     user_data)
+{
+  GromitData *data = (GromitData *) user_data;
+  undo_drawing (data);
+}
+
+void on_redo(GtkMenuItem *menuitem,
+	     gpointer     user_data)
+{
+  GromitData *data = (GromitData *) user_data;
+  redo_drawing (data);
+}
 
 
-  /* Attach the callback functions to the respective activate signal */
-  // TODO add per-device submenu for toggle_paint_item 
-  g_signal_connect(G_OBJECT (clear_item), "activate",
-		   G_CALLBACK (on_clear),
-		   data);
-  g_signal_connect(G_OBJECT (toggle_vis_item), "activate",
-		   G_CALLBACK (on_toggle_vis),
-		   data);
+void on_help(GtkMenuItem *menuitem,
+	     gpointer     user_data)
+{
+    GromitData *data = (GromitData *) user_data;
 
-  g_signal_connect(G_OBJECT (thicker_lines_item), "activate",
-		   G_CALLBACK (on_thicker_lines),
-		   data);
-  g_signal_connect(G_OBJECT (thinner_lines_item), "activate",
-		   G_CALLBACK (on_thinner_lines),
-		   data);
- 
+    gchar helpString[4096];
+    snprintf(helpString, 4096,  "The available commands are:\n\n<tt><b>\
+    toggle painting:         %s\n\
+    clear screen:            SHIFT-%s\n\
+    toggle visibility:       CTRL-%s\n\
+    quit:                    ALT-%s\n\
+    undo last stroke:        %s\n\
+    redo last undone stroke: SHIFT-%s</b></tt>",
+	     data->hot_keyval, data->hot_keyval, data->hot_keyval, data->hot_keyval,
+	     data->undo_keyval, data->undo_keyval);
 
-  /* We do need to show menu items */
-  gtk_widget_show (toggle_paint_item);
-  gtk_widget_show (clear_item);
-  gtk_widget_show (toggle_vis_item);
-  gtk_widget_show (thicker_lines_item);
-  gtk_widget_show (thinner_lines_item);
- 
+    GtkWidget *dialog = gtk_message_dialog_new(NULL,
+					       GTK_DIALOG_DESTROY_WITH_PARENT,
+					       GTK_MESSAGE_INFO,
+					       GTK_BUTTONS_CLOSE,
+					       NULL);
+    gtk_message_dialog_set_markup (GTK_MESSAGE_DIALOG (dialog), helpString);
 
-  /* show menu */
-  gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
-                  0, gtk_get_current_event_time());
+    g_signal_connect_swapped (dialog, "response",
+			      G_CALLBACK (gtk_widget_destroy),
+			      dialog);
+
+    gtk_widget_show_all(dialog);
 }
 
 
 
-
-void on_trayicon_menu (GtkStatusIcon *status_icon,
-		       guint          button,
-		       guint          activate_time,
-		       gpointer       user_data)
+void on_about(GtkMenuItem *menuitem,
+	      gpointer     user_data)
 {
-  GromitData *data = (GromitData *) user_data;
-  if(data->debug)
-    g_printerr("DEBUG: trayicon menu popup\n");
-
-  /* create the menu */
-  GtkWidget *menu = gtk_menu_new ();
-  /* Create the menu items */
-  //TODO option menu
-  GtkWidget* sep_item = gtk_separator_menu_item_new();
-  GtkWidget* quit_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, NULL);
-
-
-  /* Add them to the menu */
-  gtk_menu_shell_append (GTK_MENU_SHELL (menu), sep_item);
-  gtk_menu_shell_append (GTK_MENU_SHELL (menu), quit_item);
-
-  /* Attach the callback functions to the respective activate signal */
-  g_signal_connect(G_OBJECT (quit_item), "activate",
-		   G_CALLBACK (gtk_main_quit),
-		   NULL);
-
-
-  /* We do need to show menu items */
-  gtk_widget_show (sep_item);
-  gtk_widget_show (quit_item);
-
-
-  /* show menu */
-  gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
-                  0, gtk_get_current_event_time());
+    const gchar *authors [] = { "Christian Beier <dontmind@freeshell.org>", "Simon Budig <Simon.Budig@unix-ag.org>", NULL };
+    gtk_show_about_dialog (NULL,
+			   "program-name", "Gromit-MPX",
+			   "logo-icon-name", "gromit-mpx",
+			   "title", "About Gromit-MPX",
+			   "comments", "Gromit-MPX (GRaphics Over MIscellaneous Things) is a small tool to make annotations on the screen. Gromit-MPX is a multi-pointer port of the original Gromit annotation tool by Simon Budig.",
+			   "version", VERSION,
+			   "website", "https://github.com/bk138/gromit-mpx",
+			   "authors", authors,
+			   "copyright", "Copyright 2000 Simon Budig, 2009-2016 Christian Beier",
+			   "license-type", GTK_LICENSE_GPL_2_0,
+			   NULL);
 }
 
 
-
