@@ -24,12 +24,12 @@
 #include <string.h>
 #include <math.h>
 #include <stdlib.h>
-#include <X11/extensions/XInput2.h>
 
 #include "callbacks.h"
 #include "config.h"
 #include "input.h"
 #include "gromit-mpx.h"
+#include "build-config.h"
 
 #include "paint_cursor.xpm"
 #include "erase_cursor.xpm"
@@ -38,9 +38,10 @@
 
 GromitPaintContext *paint_context_new (GromitData *data, 
 				       GromitPaintType type,
-				       GdkColor *paint_color, 
+				       GdkRGBA *paint_color, 
 				       guint width,
-				       guint arrowsize)
+				       guint arrowsize,
+				       guint minwidth)
 {
   GromitPaintContext *context;
 
@@ -49,12 +50,13 @@ GromitPaintContext *paint_context_new (GromitData *data,
   context->type = type;
   context->width = width;
   context->arrowsize = arrowsize;
+  context->minwidth = minwidth;
   context->paint_color = paint_color;
 
   
   context->paint_ctx = cairo_create (data->backbuffer);
 
-  gdk_cairo_set_source_color(context->paint_ctx, paint_color);
+  gdk_cairo_set_source_rgba(context->paint_ctx, paint_color);
   if(!data->composited)
     cairo_set_antialias(context->paint_ctx, CAIRO_ANTIALIAS_NONE);
   cairo_set_line_width(context->paint_ctx, width);
@@ -90,9 +92,9 @@ void paint_context_print (gchar *name,
   }
 
   g_printerr ("width: %3d, ", context->width);
+  g_printerr ("minwidth: %3d, ", context->minwidth);
   g_printerr ("arrowsize: %.2f, ", context->arrowsize);
-  g_printerr ("color: #%02X%02X%02X\n", context->paint_color->red >> 8,
-              context->paint_color->green >> 8, context->paint_color->blue >> 8);
+  g_printerr ("color: %s\n", gdk_rgba_to_string(context->paint_color));
 }
 
 
@@ -434,6 +436,98 @@ void select_tool (GromitData *data,
 }
 
 
+
+void snap_undo_state (GromitData *data)
+{
+  if(data->debug)
+    g_printerr ("DEBUG: Snapping undo buffer %d.\n", data->undo_head);
+
+  copy_surface(data->undobuffer[data->undo_head], data->backbuffer);
+
+  // Increment head position
+  data->undo_head++;
+  if(data->undo_head >= GROMIT_MAX_UNDO)
+    data->undo_head -= GROMIT_MAX_UNDO;
+  data->undo_depth++;
+  // See if we ran out of undo levels with oldest undo overwritten
+  if(data->undo_depth > GROMIT_MAX_UNDO)
+    data->undo_depth = GROMIT_MAX_UNDO;
+  // Invalidate any redo from this position
+  data->redo_depth = 0;
+}
+
+
+
+void copy_surface (cairo_surface_t *dst, cairo_surface_t *src)
+{
+  cairo_t *cr = cairo_create(dst);
+  cairo_set_source_surface(cr, src, 0, 0);
+  cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
+  cairo_paint (cr);
+  cairo_destroy(cr);
+}
+
+
+
+void swap_surfaces (cairo_surface_t *a, cairo_surface_t *b)
+{
+  int width = cairo_image_surface_get_width(a);
+  int height = cairo_image_surface_get_height(a);
+  cairo_surface_t *temp = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
+  copy_surface(temp, a);
+  copy_surface(a, b);
+  copy_surface(b, temp);
+  cairo_surface_destroy(temp);
+}
+
+
+
+void undo_drawing (GromitData *data)
+{
+  /* Swap undobuffer[head-1]->backbuffer */
+  if(data->undo_depth <= 0)
+    return;
+  data->undo_depth--;
+  data->redo_depth++;
+  if(data->redo_depth > GROMIT_MAX_UNDO)
+    data->redo_depth -= GROMIT_MAX_UNDO;
+  data->undo_head--;
+  if(data->undo_head < 0)
+    data->undo_head += GROMIT_MAX_UNDO;
+
+  swap_surfaces(data->backbuffer, data->undobuffer[data->undo_head]);
+
+  GdkRectangle rect = {0, 0, data->width, data->height};
+  gdk_window_invalidate_rect(gtk_widget_get_window(data->win), &rect, 0); 
+
+  if(data->debug)
+    g_printerr ("DEBUG: Undo drawing %d.\n", data->undo_head);
+}
+
+
+
+void redo_drawing (GromitData *data)
+{
+  if(data->redo_depth <= 0)
+    return;
+
+  swap_surfaces(data->backbuffer, data->undobuffer[data->undo_head]);
+
+  data->redo_depth--;
+  data->undo_depth++;
+  data->undo_head++;
+  if(data->undo_head >= GROMIT_MAX_UNDO)
+    data->undo_head -= GROMIT_MAX_UNDO;
+  
+  GdkRectangle rect = {0, 0, data->width, data->height};
+  gdk_window_invalidate_rect(gtk_widget_get_window(data->win), &rect, 0); 
+
+  if(data->debug)
+    g_printerr("DEBUG: Redo drawing.\n");
+}
+
+
+
 void draw_line (GromitData *data,
 		GdkDevice *dev,
 		gint x1, gint y1,
@@ -517,7 +611,7 @@ void draw_arrow (GromitData *data,
       cairo_line_to(devdata->cur_context->paint_ctx, arrowhead[3].x, arrowhead[3].y);
       cairo_fill(devdata->cur_context->paint_ctx);
 
-      gdk_cairo_set_source_color(devdata->cur_context->paint_ctx, data->black);
+      gdk_cairo_set_source_rgba(devdata->cur_context->paint_ctx, data->black);
 
       cairo_move_to(devdata->cur_context->paint_ctx, arrowhead[0].x, arrowhead[0].y);
       cairo_line_to(devdata->cur_context->paint_ctx, arrowhead[1].x, arrowhead[1].y);
@@ -526,7 +620,7 @@ void draw_arrow (GromitData *data,
       cairo_line_to(devdata->cur_context->paint_ctx, arrowhead[0].x, arrowhead[0].y);
       cairo_stroke(devdata->cur_context->paint_ctx);
 
-      gdk_cairo_set_source_color(devdata->cur_context->paint_ctx, devdata->cur_context->paint_color);
+      gdk_cairo_set_source_rgba(devdata->cur_context->paint_ctx, devdata->cur_context->paint_color);
     
       data->modified = 1;
 
@@ -544,76 +638,14 @@ void draw_arrow (GromitData *data,
  */
 
 
-/* Keyboard control */
-
-gint key_press_event (GtkWidget   *grab_widget,
-		      GdkEventKey *event,
-		      gpointer     func_data)
-{
-  GromitData *data = (GromitData *) func_data;
-  GdkDevice *dev = gdk_event_get_device((GdkEvent*)event);
-
-  if (event->type == GDK_KEY_PRESS &&
-      event->hardware_keycode == data->hot_keycode)
-    {
-      if(data->debug)
-	g_printerr("DEBUG: Received hotkey press from device '%s'\n", gdk_device_get_name(dev));
-
-      if (event->state & GDK_SHIFT_MASK)
-        clear_screen (data);
-      else if (event->state & GDK_CONTROL_MASK)
-        toggle_visibility (data);
-      else if (event->state & GDK_MOD1_MASK)
-        gtk_main_quit ();
-      else
-	{
-	  /* GAAAH */
-	  gint kbd_dev_id = -1;
-	  g_object_get(dev, "device-id", &kbd_dev_id, NULL);
-
-	  XIDeviceInfo* devinfo;
-	  int devicecount = 0;
-	  gint ptr_dev_id = -1;
-	  
-	  devinfo = XIQueryDevice(GDK_DISPLAY_XDISPLAY(data->display),
-				  kbd_dev_id,
-				  &devicecount);
-	  if(devicecount)
-	    ptr_dev_id = devinfo->attachment;
-	  XIFreeDeviceInfo(devinfo);
-	  
- 
-	  GdkDeviceManager *device_manager = gdk_display_get_device_manager(data->display);
-	  GList *devices, *d;
-	  gint some_dev_id;
-	  GdkDevice *some_device = NULL;
-	  devices = gdk_device_manager_list_devices(device_manager, GDK_DEVICE_TYPE_MASTER);
-	  for(d = devices; d; d = d->next)
-	    {
-	      some_device = (GdkDevice *) d->data;
-	      g_object_get(some_device, "device-id", &some_dev_id, NULL);
-	      if(some_dev_id == ptr_dev_id)
-		break;
-	    }
-
-
-	  toggle_grab(data, some_device);
-	}
-
-      return TRUE;
-    }
-  return FALSE;
-}
-
-
 void main_do_event (GdkEventAny *event,
 		    GromitData  *data)
 {
-
+  guint keycode = ((GdkEventKey *) event)->hardware_keycode;
   if ((event->type == GDK_KEY_PRESS ||
        event->type == GDK_KEY_RELEASE) &&
       event->window == data->root &&
-      ((GdkEventKey *) event)->hardware_keycode == data->hot_keycode)
+      (keycode == data->hot_keycode || keycode == data->undo_keycode))
     {
       /* redirect the event to our main window, so that GTK+ doesn't
        * throw it away (there is no GtkWidget for the root window...)
@@ -635,12 +667,12 @@ void setup_main_app (GromitData *data, gboolean activate)
   g_free(data->white);
   g_free(data->black);
   g_free(data->red);
-  data->white = g_malloc (sizeof (GdkColor));
-  data->black = g_malloc (sizeof (GdkColor));
-  data->red   = g_malloc (sizeof (GdkColor));
-  gdk_color_parse ("#FFFFFF", data->white);
-  gdk_color_parse ("#000000", data->black);
-  gdk_color_parse ("#FF0000", data->red);
+  data->white = g_malloc (sizeof (GdkRGBA));
+  data->black = g_malloc (sizeof (GdkRGBA));
+  data->red   = g_malloc (sizeof (GdkRGBA));
+  gdk_rgba_parse(data->white, "#FFFFFF");
+  gdk_rgba_parse(data->black, "#000000");
+  gdk_rgba_parse(data->red, "#FF0000");
 
 
   /* 
@@ -666,7 +698,21 @@ void setup_main_app (GromitData *data, gboolean activate)
   */
   /* SHAPE SURFACE*/
   cairo_surface_destroy(data->backbuffer);
-  data->backbuffer = cairo_image_surface_create(CAIRO_FORMAT_ARGB32 ,data->width, data->height);
+  data->backbuffer = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, data->width, data->height);
+  
+  /*
+    UNDO STATE
+  */
+  data->undo_head = 0;
+  data->undo_depth = 0;
+  data->redo_depth = 0;
+  int i;
+  for (i = 0; i < GROMIT_MAX_UNDO; i++)
+    {
+      cairo_surface_destroy(data->undobuffer[i]);
+      data->undobuffer[i] = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, data->width, data->height);
+    }
+  
 
 
   /* EVENTS */
@@ -725,6 +771,8 @@ void setup_main_app (GromitData *data, gboolean activate)
   gtk_selection_add_target (data->win, GA_CONTROL, GA_VISIBILITY, 5);
   gtk_selection_add_target (data->win, GA_CONTROL, GA_CLEAR, 6);
   gtk_selection_add_target (data->win, GA_CONTROL, GA_RELOAD, 7);
+  gtk_selection_add_target (data->win, GA_CONTROL, GA_UNDO, 8);
+  gtk_selection_add_target (data->win, GA_CONTROL, GA_REDO, 9);
 
 
  
@@ -766,6 +814,34 @@ void setup_main_app (GromitData *data, gboolean activate)
         }
     }
 
+  /*
+     FIND UNDOKEY KEYCODE 
+  */
+  if (data->undo_keyval)
+    {
+      GdkKeymap    *keymap;
+      GdkKeymapKey *keys;
+      gint          n_keys;
+      guint         keyval;
+
+      if (strlen (data->undo_keyval) > 0 &&
+          strcasecmp (data->undo_keyval, "none") != 0)
+        {
+          keymap = gdk_keymap_get_for_display (data->display);
+          keyval = gdk_keyval_from_name (data->undo_keyval);
+
+          if (!keyval || !gdk_keymap_get_entries_for_keyval (keymap, keyval,
+                                                             &keys, &n_keys))
+            {
+              g_printerr ("cannot find the key \"%s\"\n", data->undo_keyval);
+              exit (1);
+            }
+
+          data->undo_keycode = keys[0].keycode;
+          g_free (keys);
+        }
+    }
+
 
   /* 
      INPUT DEVICES
@@ -785,14 +861,14 @@ void setup_main_app (GromitData *data, gboolean activate)
   data->modified = 0;
 
   data->default_pen = paint_context_new (data, GROMIT_PEN,
-					 data->red, 7, 0);
+					 data->red, 7, 0, 1);
   data->default_eraser = paint_context_new (data, GROMIT_ERASER,
-					    data->red, 75, 0);
+					    data->red, 75, 0, 1);
 
   
 
   gdk_event_handler_set ((GdkEventFunc) main_do_event, data, NULL);
-  gtk_key_snooper_install (key_press_event, data);
+  gtk_key_snooper_install (snoop_key_press, data);
 
   if (activate)
     acquire_grab (data, NULL); /* grab all */
@@ -800,15 +876,113 @@ void setup_main_app (GromitData *data, gboolean activate)
   /* 
      TRAY ICON
   */
-  data->trayicon = gtk_status_icon_new_from_file("/usr/share/pixmaps/gromit-mpx.png");
-  gtk_status_icon_set_tooltip_text (data->trayicon, "Gromit-MPX");
-  g_signal_connect (data->trayicon, "activate",
-		    G_CALLBACK (on_trayicon_activate), data);
-  g_signal_connect (data->trayicon, "popup-menu",
-		    G_CALLBACK (on_trayicon_menu), data);
+  data->trayicon = app_indicator_new (PACKAGE_NAME,
+				      PACKAGE_NAME,
+				      APP_INDICATOR_CATEGORY_APPLICATION_STATUS);
+
+  app_indicator_set_status (data->trayicon, APP_INDICATOR_STATUS_ACTIVE);
+
+
+
+  /* create the menu */
+  GtkWidget *menu = gtk_menu_new ();
+
+  /* Create the menu items */
+  GtkWidget* toggle_paint_item = gtk_menu_item_new_with_label ("Toggle Painting");
+  GtkWidget* clear_item = gtk_menu_item_new_with_label ("Clear Screen");
+  GtkWidget* toggle_vis_item = gtk_menu_item_new_with_label ("Toggle Visibility");
+  GtkWidget* thicker_lines_item = gtk_menu_item_new_with_label ("Thicker Lines");
+  GtkWidget* thinner_lines_item = gtk_menu_item_new_with_label ("Thinner Lines");
+  GtkWidget* opacity_bigger_item = gtk_menu_item_new_with_label ("Bigger Opacity");
+  GtkWidget* opacity_lesser_item = gtk_menu_item_new_with_label ("Lesser Opacity");
+  GtkWidget* undo_item = gtk_menu_item_new_with_label ("Undo");
+  GtkWidget* redo_item = gtk_menu_item_new_with_label ("Redo");
+
+  GtkWidget* sep_item = gtk_separator_menu_item_new();
+  GtkWidget* help_item = gtk_menu_item_new_with_mnemonic("_Help");
+  GtkWidget* about_item = gtk_menu_item_new_with_mnemonic("_About");
+  GtkWidget* quit_item = gtk_menu_item_new_with_mnemonic("_Quit");
+
+
+  /* Add them to the menu */
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), toggle_paint_item);
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), clear_item);
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), toggle_vis_item);
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), thicker_lines_item);
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), thinner_lines_item);
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), opacity_bigger_item);
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), opacity_lesser_item);
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), undo_item);
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), redo_item);
+
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), sep_item);
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), help_item);
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), about_item);
+  gtk_menu_shell_append (GTK_MENU_SHELL (menu), quit_item);
+
+
+  /* Attach the callback functions to the respective activate signal */
+  g_signal_connect(toggle_paint_item, "button-press-event",
+		   G_CALLBACK(on_toggle_paint),
+		   data);
+  g_signal_connect(G_OBJECT (clear_item), "activate",
+		   G_CALLBACK (on_clear),
+		   data);
+  g_signal_connect(G_OBJECT (toggle_vis_item), "activate",
+		   G_CALLBACK (on_toggle_vis),
+		   data);
+  g_signal_connect(G_OBJECT (thicker_lines_item), "activate",
+		   G_CALLBACK (on_thicker_lines),
+		   data);
+  g_signal_connect(G_OBJECT (thinner_lines_item), "activate",
+		   G_CALLBACK (on_thinner_lines),
+		   data);
+  g_signal_connect(G_OBJECT (opacity_bigger_item), "activate",
+		   G_CALLBACK (on_opacity_bigger),
+		   data);
+  g_signal_connect(G_OBJECT (opacity_lesser_item), "activate",
+		   G_CALLBACK (on_opacity_lesser),
+		   data);
+  g_signal_connect(G_OBJECT (undo_item), "activate",
+		   G_CALLBACK (on_undo),
+		   data);
+  g_signal_connect(G_OBJECT (redo_item), "activate",
+		   G_CALLBACK (on_redo),
+		   data);
+
+  g_signal_connect(G_OBJECT (help_item), "activate",
+		   G_CALLBACK (on_help),
+		   data);
+  g_signal_connect(G_OBJECT (about_item), "activate",
+		   G_CALLBACK (on_about),
+		   NULL);
+  g_signal_connect(G_OBJECT (quit_item), "activate",
+		   G_CALLBACK (gtk_main_quit),
+		   NULL);
+
+
+  /* We do need to show menu items */
+  gtk_widget_show (toggle_paint_item);
+  gtk_widget_show (clear_item);
+  gtk_widget_show (toggle_vis_item);
+  gtk_widget_show (thicker_lines_item);
+  gtk_widget_show (thinner_lines_item);
+  gtk_widget_show (opacity_bigger_item);
+  gtk_widget_show (opacity_lesser_item);
+  gtk_widget_show (undo_item);
+  gtk_widget_show (redo_item);
+
+  gtk_widget_show (sep_item);
+  gtk_widget_show (help_item);
+  gtk_widget_show (about_item);
+  gtk_widget_show (quit_item);
+
+
+  app_indicator_set_menu (data->trayicon, GTK_MENU(menu));
 
 }
 
+
 void parse_print_help (gpointer key, gpointer value, gpointer user_data)
 {
   paint_context_print ((gchar *) key, (GromitPaintContext *) value);
@@ -825,6 +999,9 @@ int app_parse_args (int argc, char **argv, GromitData *data)
    data->hot_keyval = DEFAULT_HOTKEY;
    data->hot_keycode = 0;
 
+   data->undo_keyval = DEFAULT_UNDOKEY;
+   data->undo_keycode = 0;
+
    for (i=1; i < argc ; i++)
      {
        arg = argv[i];
@@ -868,6 +1045,52 @@ int app_parse_args (int argc, char **argv, GromitData *data)
                wrong_arg = TRUE;
              }
          }
+      else if (strcmp (arg, "-o") == 0 ||
+                strcmp (arg, "--opacity") == 0)
+         {
+           if (i+1 < argc && strtod (argv[i+1], NULL) >= 0.0 && strtod (argv[i+1], NULL) <= 1.0)
+             {
+               data->opacity = strtod (argv[i+1], NULL);
+               g_printerr ("Opacity set to: %.2f\n", data->opacity);
+               gtk_widget_set_opacity(data->win, data->opacity);
+               i++;
+             }
+           else
+             {
+               g_printerr ("-o requires an opacity >= 0,0 && <=1,0 as argument\n");
+               wrong_arg = TRUE;
+             }
+         }
+       else if (strcmp (arg, "-u") == 0 ||
+                strcmp (arg, "--undo-key") == 0)
+         {
+           if (i+1 < argc)
+             {
+               data->undo_keyval = argv[i+1];
+               data->undo_keycode = 0;
+               i++;
+             }
+           else
+             {
+               g_printerr ("-u requires an Key-Name as argument\n");
+               wrong_arg = TRUE;
+             }
+         }
+       else if (strcmp (arg, "-U") == 0 ||
+                strcmp (arg, "--undo-keycode") == 0)
+         {
+           if (i+1 < argc && atoi (argv[i+1]) > 0)
+             {
+               data->undo_keyval = NULL;
+               data->undo_keycode = atoi (argv[i+1]);
+               i++;
+             }
+           else
+             {
+               g_printerr ("-U requires an keycode > 0 as argument\n");
+               wrong_arg = TRUE;
+             }
+         }
        else
          {
            g_printerr ("Unknown Option for Gromit-MPX startup: \"%s\"\n", arg);
@@ -937,6 +1160,16 @@ int main_client (int argc, char **argv, GromitData *data)
          {
            action = GA_RELOAD;
          }
+       else if (strcmp (arg, "-z") == 0 ||
+                strcmp (arg, "--undo") == 0)
+         {
+           action = GA_UNDO;
+         }
+       else if (strcmp (arg, "-y") == 0 ||
+                strcmp (arg, "--redo") == 0)
+         {
+           action = GA_REDO;
+         }
        else
          {
            g_printerr ("Unknown Option to control a running Gromit-MPX process: \"%s\"\n", arg);
@@ -977,6 +1210,7 @@ int main (int argc, char **argv)
   data->root = gdk_screen_get_root_window (data->screen);
   data->width = gdk_screen_get_width (data->screen);
   data->height = gdk_screen_get_height (data->screen);
+  data->opacity = DEFAULT_OPACITY;
 
   /*
     init our window
@@ -987,7 +1221,7 @@ int main (int argc, char **argv)
 
   gtk_window_fullscreen(GTK_WINDOW(data->win)); 
   gtk_window_set_skip_taskbar_hint(GTK_WINDOW(data->win), TRUE);
-  gtk_window_set_opacity(GTK_WINDOW(data->win), 0.75);
+  gtk_widget_set_opacity(data->win, data->opacity);
   gtk_widget_set_app_paintable (data->win, TRUE);
   gtk_window_set_decorated (GTK_WINDOW (data->win), FALSE);
 
